import io
import pickle
import numpy as np
import os
import time
from numba import jit
from mpi4py import MPI

import config
from utils import read_pwms_from_pmf, calculate_background, extractSequenceWorker, splitGenome
from config import *
from sharedEnums import ParallelPwmTags as pmt


@jit(nopython=True, parallel=False)
def calculate_kld(sequence: np.ndarray, pwm: np.ndarray,
                  background: np.ndarray) -> np.ndarray:  # would be nice to do a numpy version and check if numpy or numba is faster.
    """calculates the kullback leibler divergence of a dna sequence given a pwm and the background probabilities.

    Args:
        sequence (np.ndarray): dna sequence as np.array, see function sequence2nparray() in utils.py
        pwm (np.ndarray): position weight matrix, columns are position, rows are bases in order A,C,G,T
        background (np.ndarray): The background probability of the bases in the order A,C,G,T

    Returns:
        kld_values (np.ndarray): array of kld values at the respective position (length = length(sequence)-length(pwm)+1)
    """
    results: np.ndarray = np.zeros(len(sequence) - pwm.shape[1] + 1)
    for start_idx in range(len(sequence) - pwm.shape[1]):
        kld: float = 0.0
        for cur_pwm_pos in range(pwm.shape[1]):
            b = sequence[start_idx + cur_pwm_pos]
            if b == 4:
                # equal to 0, not sure if this is the correct behaviour, but shouldn't matter for the task at hand
                # kld += background[b]*np.log2(background[b]/background[b])
                continue
            kld += pwm[b, cur_pwm_pos] * np.log2(pwm[b, cur_pwm_pos] / background[b])
        results[start_idx] = kld

    return results


if __name__ == '__main__':
    """
    Main function of the calculation of the kld on a genome. 
    Please use the config file to specify the parameters of the programm (paths).
    Please use the sbatch script to change the slurm parameters (number of nodes, etc.)
    """

    comm = MPI.COMM_WORLD
    rank: int = comm.Get_rank()
    nprocs: int = comm.Get_size()

    if rank == 0:
        print(f"nprocs: {nprocs}")

        # only root process prepares everything, this could be parallelized later
        time_preparation: int = time.perf_counter_ns()

        pwms: dict = read_pwms_from_pmf(pwm_path)
        pwm_names: list = list(pwms.keys())
        number_of_pwms: int = len(pwm_names) if number_of_pwms == 0 else number_of_pwms
        pwm_names = pwm_names[:number_of_pwms]

        background: np.ndarray = calculate_background()

        # prepare sequence
        if rebuilt_files:
            print('splitting Genome, this may take a while')
            splitGenome(genome_path, workdir)

        files = os.listdir(workdir)
        # check if everything worked
        print('check if all files were created')
        with open(workdir+os.sep+'file_list.txt', 'r') as fp:
            for line in fp:
                if line.strip() not in files:
                    raise ValueError(f"file {line.strip()} does not exist in workdir. Please set rebuilt_files = True to rebuilt the files.")

        print(f'The preparation of the PWMs and the sequence took {time_preparation / 1_000_000_000} s')
    else:
        background: np.ndarray = np.empty(5, dtype=np.float64)

    # broadcast background probablility
    if rank == 0:
        print("broadcast background probabilities")
    comm.Bcast(background, root=0)

    # distribute tasks and calculate klds
    if rank == 0:
        performance_dict: dict = {} # performance values
        open_tasks: dict = {}  # all open tasks
        finished_tasks: set = set()  # all finished tasks
        current_tasks: dict = {}  # all tasks currently worked on by specific worker
        keep_working: bool = True  # signals when to stop/kill workers
        total_calculations = 0
        combined_time = 0
        worker_chromosome: dict = {}
        next_chromosome_index: int = 0
        sorted_chromosome_files: list = []

        with open(workdir+os.sep+"file_list.txt", 'r') as fp:
            for line in fp:
                sorted_chromosome_files.append(line.replace("\n", ""))

        for chromosome in sorted_chromosome_files:
            open_tasks[chromosome] = pwm_names

        for worker in range(1, nprocs):
            worker_chromosome[worker] = []

        # print(f"sorted_chromosome_files {len(sorted_chromosome_files)}")
        # print(f"pwm_names {len(pwm_names)}")

        n_tasks: int = len(sorted_chromosome_files)*len(pwm_names)
        calculations_finished = 0  # to have a nicer output

        def selectNextChromosome(next_chromosome_index) -> int:
            for chr_idx in range(next_chromosome_index, next_chromosome_index + len(sorted_chromosome_files)):
                chr_idx = chr_idx % len(sorted_chromosome_files)
                if len(open_tasks[sorted_chromosome_files[chr_idx]]) > 0:
                    return chr_idx
            return -1

        while len(finished_tasks) < n_tasks:
            # check if workers need work
            for worker in range(1, nprocs):  # loop through workers
                if comm.Iprobe(source=worker, tag=pmt.CALC_DONE):  # check if worker is done with its job (and download results)
                    comm.recv(source=worker, tag=pmt.CALC_DONE)  # clean message
                    tmp_task = current_tasks[worker]  # get current task of worker

                    tmp_time = comm.recv(source=worker, tag=pmt.CALC_TIME)  # get calculation time
                    # tmp_performance = comm.recv(source=worker, tag=pmt.PERFORMANCE)
                    combined_time += tmp_time

                    # bookkeeping
                    finished_tasks.add(tmp_task)
                    current_tasks[worker] = None

                    # calculations_finished += tmp_task[2]
                    # if len(finished_tasks) % (nprocs//2) == 0:
                    #     time_seconds = combined_time // 10**9
                    #     print(f"calculations: {round(calculations_finished/total_calculations*100, 2)} %, tasks: {round(len(finished_tasks)/n_tasks*100, 2)} %, "
                    #           f"combined computation time: {time_seconds//60//60} h {(time_seconds//60)-(time_seconds//60//60)} min "
                    #           f"{time_seconds%60} s                                           \n", end='\r')

                if comm.Iprobe(source=worker, tag=pmt.IDLE) and len(open_tasks) > 0:  # if worker is idle, give new task
                    comm.recv(source=worker, tag=pmt.IDLE)

                    # check if there is still work and assign new chromosome if needed.

                    if keep_working:
                        if len(worker_chromosome[worker]) == 0:
                            worker_chromosome[worker] = [selectNextChromosome(next_chromosome_index)]
                            # print("line 155", worker, next_chromosome_index, worker_chromosome[worker])
                            next_chromosome_index = worker_chromosome[worker][-1] + 1
                        elif len(open_tasks[sorted_chromosome_files[worker_chromosome[worker][-1]]]) == 0:
                            worker_chromosome[worker] = worker_chromosome[worker] + [selectNextChromosome(next_chromosome_index)]
                            next_chromosome_index = worker_chromosome[worker][-1] + 1
                            # print("line 160", worker, next_chromosome_index, worker_chromosome[worker])

                    # for some reason we had a bug here, we thought the first condition should suffice, but there is somewhere an error.
                    # due to time reasons we will not hunt the bug.
                    if worker_chromosome[worker][-1] == -1 or len(open_tasks[sorted_chromosome_files[worker_chromosome[worker][-1]]])==0:
                        keep_working = False
                        comm.send(keep_working, dest=worker, tag=pmt.KEEP_WORKING)
                        continue

                    comm.send(keep_working, dest=worker, tag=pmt.KEEP_WORKING)

                    # get task
                    if len(open_tasks[sorted_chromosome_files[worker_chromosome[worker][-1]]]) > 0:  # if there is still work on the last chromosome
                        tmp_pwm = open_tasks[sorted_chromosome_files[worker_chromosome[worker][-1]]][-1]
                        open_tasks[sorted_chromosome_files[worker_chromosome[worker][-1]]] = open_tasks[sorted_chromosome_files[worker_chromosome[worker][-1]]][:-1]  # remove last pwm

                        tmp_chromosome = sorted_chromosome_files[worker_chromosome[worker][-1]]
                        tmp_task = (tmp_chromosome, tmp_pwm)

                        current_tasks[worker] = tmp_task
                    else:


                        raise ValueError("This should NEVER happen")

                    # send info, so worker can start calculating
                    comm.send(pwms[tmp_task[1]]['pwm'].shape, dest=worker, tag=pmt.PWM_SHAPE)
                    comm.send(tmp_task[1], dest=worker, tag=pmt.PWM_NAME)
                    comm.send(tmp_task[0], dest=worker, tag=pmt.CHROMOSOME)
                    comm.Send(pwms[tmp_task[1]]['pwm'], dest=worker, tag=pmt.PWM)



                elif len(open_tasks) == 0:  # if no tasks left, do nothing
                    pass

        print(f"the combined calculation time is {combined_time / 1_000_000_000} s")
        print("shutting down workers")
        keep_working = False
        still_alive_workers = 0
        for worker in range(1, nprocs):
            if comm.Iprobe(source=worker, tag=pmt.IDLE):
                comm.recv(source=worker, tag=pmt.IDLE)  # clean idleing message
                comm.send(keep_working, dest=worker, tag=pmt.KEEP_WORKING)  # send stop working message
                still_alive_workers -= 1
            else:
                print(f"worker {worker} is not idle, but will shut down at the next possibility")
            comm.isend(None, dest=worker, tag=pmt.EXIT)  # this is a relict, just for safety we will leave it here

        if still_alive_workers == 0:
            print("all workers could be reached and should shut down on their own.")
        elif len(finished_tasks) == n_tasks and any(v is not None for v in current_tasks.values()):
            print("there was a problem, not all workers could be reached. One should investigate here, probably a worker was not idleing.")
            print(f"len(finished_tasks): {len(finished_tasks)}, ntasks: {n_tasks}")
            print(current_tasks)



    else:
        time_kld_start: int = time.perf_counter_ns()
        pwm_shape: tuple
        chromosome: str = ""
        currentChromosome: str = ""
        worker: int = rank
        keep_working: bool = True

        while not comm.Iprobe(source=0, tag=pmt.EXIT):
            # print(f"worker {worker} is idle")
            comm.send(None, dest=0, tag=pmt.IDLE)  # signal idleing
            keep_working = comm.recv(source=0, tag=pmt.KEEP_WORKING)
            if not keep_working:  # All work is done, the worker can now rest in peace :-)
                break

            pwm_shape: tuple = comm.recv(source=0, tag=pmt.PWM_SHAPE)
            pwm_name: str = comm.recv(source=0, tag=pmt.PWM_NAME)
            chromosome: str = comm.recv(source=0, tag=pmt.CHROMOSOME)
            tmp_pwm: np.ndarray = np.empty(pwm_shape)  # initialize pwm



            # get full chromosome object if not already known
            if not chromosome == currentChromosome:

                pickle_path = workdir+os.sep+chromosome.split('.')[0] + ".pickle"

                if(os.path.exists(pickle_path)):
                    with open(pickle_path, "rb") as fp:
                        chromosomeData = pickle.load(fp)
                else:
                    chromosomeData = extractSequenceWorker(workdir+os.sep+chromosome)
                    with open(pickle_path, "wb") as fp:
                        pickle.dump(chromosomeData, fp, protocol=pickle.HIGHEST_PROTOCOL)

                currentChromosome = chromosome


            print(f"worker {worker} is calculating chromosome {currentChromosome}")

            start_time: int = time.perf_counter_ns()
            kld_array: np.ndarray = calculate_kld(sequence=chromosomeData['sequence'], pwm=tmp_pwm, background=background)
            calc_time: int = time.perf_counter_ns() - start_time
            comm.isend(None, dest=0, tag=pmt.CALC_DONE)  # signal done.
            comm.send(len(kld_array), dest=0, tag=pmt.LENGTH_KLD)

            # print("calculation done, saving results")
            # Save results

            with open(workdir+os.sep+'results_' + pwm_name.replace('.', '_') + '.csv', 'wb') as fp:
                pickle.dump(kld_array, fp, protocol=pickle.HIGHEST_PROTOCOL)

            # np.savetxt(workdir+os.sep+'results_' + pwm_name.replace('.', '_') + '.csv', kld_array) #was really slow.

            comm.send(calc_time, dest=0, tag=pmt.CALC_TIME)  # send calculation time
