import pickle
import numpy as np
import os
import time
from numba import jit
from mpi4py import MPI

from utils_old import read_pwms_from_pmf, calculate_background, extractSequences
from config import *
from sharedEnums import ParallelPwmTags as pmt

@jit(nopython=True, parallel=False)
def calculate_kld(sequence: np.ndarray, pwm: np.ndarray,
                  background: np.ndarray) -> np.ndarray:  # would be nice to do a numpy version and check if numpy or numba is faster.
    """calculates the kullback leibler divergence of a dna sequence given a pwm and the background probabilities.

    Args:
        sequence (np.ndarray): dna sequence as np.array, see function sequence2nparray() in utils.py
        pwm (np.ndarray): position weight matrix, columns are position, rows are bases in order A,C,G,T
        background (np.ndarray): The background probability of the bases in the order A,C,G,T

    Returns:
        kld_values (np.ndarray): array of kld values at the respective position (length = length(sequence)-length(pwm)+1)
    """
    results = np.zeros(len(sequence) - pwm.shape[1] + 1)
    for start_idx in range(len(sequence) - pwm.shape[1]):
        kld = 0
        for cur_pwm_pos in range(pwm.shape[1]):
            b = sequence[start_idx + cur_pwm_pos]
            if b == 4:
                # equal to 0, not sure if this is the correct behaviour, but shouldn't matter for the task at hand
                # kld += background[b]*np.log2(background[b]/background[b])
                continue
            kld += pwm[b, cur_pwm_pos] * np.log2(pwm[b, cur_pwm_pos] / background[b])
        results[start_idx] = kld

    return results


if __name__ == '__main__':
    """
    Main function of the calculation of the kld on a genome. 
    Please use the config file to specify the parameters of the programm (paths).
    Please use the sbatch script to change the slurm parameters (number of nodes, etc.)
    """

    comm = MPI.COMM_WORLD
    rank: int = comm.Get_rank()
    nprocs: int = comm.Get_size()

    if rank == 0:
        print(f"nprocs: {nprocs}")

        # only root process prepares everything, this could be parallelized later
        time_preparation: int = time.perf_counter_ns()

        pwms: dict = read_pwms_from_pmf(pwm_path)
        pwm_names: list = list(pwms.keys())
        number_of_pwms: int = len(pwm_names) if number_of_pwms == 0 else number_of_pwms
        pwm_names = pwm_names[:number_of_pwms]

        background: np.ndarray = calculate_background()

        # prepare sequence
        if dev:  # use chromosome 21 for development
            genome: dict = extractSequences(chromosome21_path)

        elif os.path.exists(compressed_genome_path) and use_pickeled:
            with open(compressed_genome_path, 'rb') as fp:
                genome: dict = pickle.load(fp)
        else:
            genome: dict = extractSequences(genome_path)
            if use_pickeled:
                with open(compressed_genome_path, 'wb') as fp:
                    pickle.dump(genome, fp, protocol=pickle.HIGHEST_PROTOCOL)
        time_preparation: int = time.perf_counter_ns() - time_preparation

        print(f'The preparation of the PWMs and the sequence took {time_preparation / 1_000_000_000} s')
    else:
        background: np.ndarray = np.empty(5, dtype=np.float64)
        genome: dict = {}

    # broadcast PWMs and background probablility
    if rank == 0:
        time_broadcasting = time.perf_counter_ns()
    comm.Bcast(background, root=0)
    genome: dict = comm.bcast(genome, root=0)
    if rank == 0:
        time_broadcasting = time.perf_counter_ns()-time_broadcasting
        print(f"broadcasting the genome and background probabilities took {time_broadcasting / 1_000_000_000} s")

    # distribute pwms and calculate klds
    if rank == 0:
        kld: dict = {}  # used to store the results
        open_tasks: set = set()  # all open tasks
        finished_tasks: set = set()  # all finished tasks
        current_tasks: dict = {}  # all tasks currently worked on by specific worker
        keep_working: bool = True  # signals when to stop/kill workers
        total_calculations = 0
        combined_time = 0

        for chromosome in genome.keys():
            kld[chromosome] = {}
            len_current_chromosome = len(genome[chromosome])
            for name in pwm_names:
                kld[chromosome][name] = {}
                calculations = len_current_chromosome * pwms[name]['pwm'].shape[1]
                open_tasks.add((chromosome, name, calculations))
                total_calculations += calculations

        n_tasks: int = len(open_tasks)
        calculations_finished = 0  # to have a nicer output
        while len(finished_tasks) < n_tasks:
            # check if workers need work
            for worker in range(1, nprocs):  # loop through workers
                if comm.Iprobe(source=worker, tag=pmt.CALC_DONE):  # check if worker is done with its job (and download results)
                    comm.recv(source=worker, tag=pmt.CALC_DONE)  # clean message
                    tmp_task = current_tasks[worker]  # get current task of worker

                    shape_kld = comm.recv(source=worker, tag=pmt.LENGTH_KLD)  # get the length of the result to initiate array
                    kld[tmp_task[0]][tmp_task[1]]['kld'] = np.empty(shape_kld)  # initiate result array
                    comm.Recv(kld[tmp_task[0]][tmp_task[1]]['kld'], source=worker, tag=pmt.KLD)  # get result kld

                    tmp_time = comm.recv(source=worker, tag=pmt.CALC_TIME)  # get calculation time
                    kld[tmp_task[0]][tmp_task[1]]['time'] = tmp_time
                    kld[tmp_task[0]][tmp_task[1]]['performance'] = tmp_task[2]/tmp_time*10**9  # operations per second
                    combined_time += tmp_time

                    # bookkeeping
                    finished_tasks.add(tmp_task)
                    current_tasks[worker] = None

                    calculations_finished += tmp_task[2]
                    if len(finished_tasks) % (nprocs//2) == 0:
                        time_seconds = combined_time // 10**9
                        print(f"calculations: {round(calculations_finished/total_calculations*100, 2)} %, tasks: {round(len(finished_tasks)/n_tasks*100, 2)} %, "
                              f"combined computation time: {time_seconds//60//60} h {(time_seconds//60)-(time_seconds//60//60)} min "
                              f"{time_seconds%60} s                                           ", end='\r')

                if comm.Iprobe(source=worker, tag=pmt.IDLE) and len(open_tasks) > 0:  # if worker is idle, give new task
                    comm.recv(source=worker, tag=pmt.IDLE)
                    comm.send(keep_working, dest=worker, tag=pmt.KEEP_WORKING)

                    # get task
                    tmp_task = open_tasks.pop()
                    current_tasks[worker] = tmp_task

                    # send info, so worker can start calculating
                    comm.send(pwms[tmp_task[1]]['pwm'].shape, dest=worker, tag=pmt.PWM_SHAPE)
                    comm.send(tmp_task[0], dest=worker, tag=pmt.CHROMOSOME)
                    comm.Send(pwms[tmp_task[1]]['pwm'], dest=worker, tag=pmt.PWM)

                elif len(open_tasks) == 0:  # if no tasks left, do nothing
                    pass

        print("shutting down workers")
        keep_working = False
        still_alive_workers = 0
        for worker in range(1, nprocs):
            if comm.Iprobe(source=worker, tag=pmt.IDLE):
                comm.recv(source=worker, tag=pmt.IDLE)  # clean idleing message
                comm.send(keep_working, dest=worker, tag=pmt.KEEP_WORKING)  # send stop working message
                still_alive_workers -= 1
            else:
                print(f"worker {worker} is not idle")
            comm.isend(None, dest=worker, tag=pmt.EXIT)  # this is a relict, just for safety we will leave it here

        if still_alive_workers == 0:
            print("all workers could be reached and should shut down on their own.")
        elif len(finished_tasks) == n_tasks and any(v is not None for v in current_tasks.values()):
            print("there was a problem, not all workers could be reached. One should investigate here, probably a worker was not idleing.")
            print(f"len(finished_tasks): {len(finished_tasks)}, ntasks: {n_tasks}")
            print(current_tasks)



    else:
        time_kld_start: int = time.perf_counter_ns()
        pwm_shape: tuple
        chromosome: str = ""
        worker: int = rank
        keep_working: bool = True
        while not comm.Iprobe(source=0, tag=pmt.EXIT):
            # print(f"worker {worker} is idle")
            comm.send(None, dest=0, tag=pmt.IDLE)  # signal idleing
            keep_working = comm.recv(source=0, tag=pmt.KEEP_WORKING)
            if not keep_working:  # All work is done, the worker can now rest in peace :-)
                break

            pwm_shape: tuple = comm.recv(source=0, tag=pmt.PWM_SHAPE)
            chromosome: str = comm.recv(source=0, tag=pmt.CHROMOSOME)
            tmp_pwm: np.ndarray = np.empty(pwm_shape)  # initialize pwm
            comm.Recv(tmp_pwm, source=0, tag=pmt.PWM)

            # print(f"worker {worker} is calculating")
            start_time: int = time.perf_counter_ns()
            kld_array: np.ndarray = calculate_kld(sequence=genome[chromosome], pwm=tmp_pwm, background=background)
            calc_time: int = time.perf_counter_ns() - start_time
            comm.isend(None, dest=0, tag=pmt.CALC_DONE)  # signal done.
            comm.send(len(kld_array), dest=0, tag=pmt.LENGTH_KLD)
            comm.Send(kld_array, dest=0, tag=pmt.KLD)  # send results back
            comm.send(calc_time, dest=0, tag=pmt.CALC_TIME)  # send calculation time


    # save results
    if rank == 0:
        if saving_results:
            print(f"saving result to disk")
            time_saving = time.perf_counter_ns()
            with open(compressed_output_path, "wb") as fp:
                pickle.dump(kld, fp, protocol=pickle.HIGHEST_PROTOCOL)
            time_saving = time.perf_counter_ns() - time_saving
            print(f"saving/pickeling of the output dict took {time_saving / 1_000_000_000} s")
