import pickle
import numpy as np
import os
import time
from numba import jit

from utils_old import read_pwms_from_pmf, calculate_background, extractSequences
from config import *

@jit(nopython=True, parallel=False)
def calculate_kld(sequence: np.ndarray, pwm: np.ndarray,
                  background: np.ndarray) -> np.ndarray:  # would be nice to do this in numpy and in parallel
    """calculates the kullback leibler divergence of a dna sequence given a pwm and the background probabilities.

    Args:
        sequence (str): dna sequence as np.array, see function sequence2nparray
        pwm (np.ndarray): position weight matrix, columns are position, rows are bases in order A,C,G,T
        background (np.ndarray): The background probability of the bases in the order A,C,G,T

    Returns:
        time_kld (float): time of the actual calculation in ns
        time_kld_per_thousand (float): time needed to calculate kld for 1000 positions
        kld_values (np.ndarray): array of kld values at the respective position (length = length(sequence)-length(pwm)+1)  
    """

    results = np.zeros(len(sequence) - len(pwm) +1)
    for start_idx in range(len(sequence) - len(pwm) + 1):
        kld = 0
        for cur_pwm_pos in range(len(pwm)):
            b = sequence[start_idx + cur_pwm_pos]
            if b == 4:
                # kld += background[b]*np.log2(background[b]/background[b]) #equal to 0, not sure if this is the correct behaviour
                continue
            kld += pwm[b, cur_pwm_pos] * np.log2(pwm[b, cur_pwm_pos] / background[b])
        results[start_idx] = kld

    return results


if __name__ == '__main__':

    time_preparation = time.perf_counter_ns()
    pwms = read_pwms_from_pmf(pwm_path)

    pwm_names = list(pwms.keys())
    background = calculate_background()

    # prepare sequence
    if dev: #use chrmosome 21 for development
        genome = extractSequences(chromosome21_path)

    elif os.path.exists(compressed_genome_path) and use_pickeled:
        with open(compressed_genome_path, 'rb') as fp:
            genome = pickle.load(fp)
    else:
        genome = extractSequences(genome_path)
        if use_pickeled:
            with open(compressed_genome_path, 'wb') as fp:
                pickle.dump(genome, fp, protocol=pickle.HIGHEST_PROTOCOL)
    time_preparation = time.perf_counter_ns()-time_preparation

    print(f'The preparation of the PWMs and the sequence took {time_preparation/1_000_000_000} s')


    # calculate kld
    kld = {}
    calc_time = 0
    for chr in genome.keys():
        kld[chr] = {}
        kld[chr]['kld'] = {}
        kld[chr]['time_kld'] = {}
        kld[chr]['time_kld_per_thousand_operations'] = {}
        kld[chr]['operations'] = {}
        number_of_pwms = len(pwm_names) if number_of_pwms == 0 else number_of_pwms
        for i in range(number_of_pwms):
            time_kld_start = time.perf_counter_ns()
            kld[chr]['kld'][pwm_names[i]] = calculate_kld(sequence=genome[chr], pwm=pwms[pwm_names[i]]['pwm'], background=background)
            time_kld = time.perf_counter_ns() - time_kld_start
            kld[chr]['time_kld'][pwm_names[i]] = time_kld
            kld[chr]['time_kld_per_thousand_operations'][pwm_names[i]] = time_kld * 1000 / (len(genome[chr])*pwms[pwm_names[i]]['pwm'].shape[1])
            kld[chr]['operations'][pwm_names[i]] = len(genome[chr]) * pwms[pwm_names[i]]['pwm'].shape[1]
            calc_time += time_kld
            print(f"the calculation of {chr} and {pwm_names[i]} took {time_kld} ns and averaged {kld[chr]['time_kld_per_thousand_operations'][pwm_names[i]]}")
    print(f"The calculation took {calc_time} ns ({calc_time / 1_000_000_000} s)")

    # write output
    time_saving = time.perf_counter_ns()
    with open(compressed_output_path, "wb") as fp:
        pickle.dump(kld, fp, protocol=pickle.HIGHEST_PROTOCOL)
    time_saving = time.perf_counter_ns() - time_saving
    print(f"saving/pickeling of the output dict took {time_saving / 1_000_000_000} s")

    timing = 'chromosome, "overall time(ns)", "time per thousand operations", operations\n'
    for chr in genome.keys():
        timing += f"\"{chr}\", {list(kld[chr]['time_kld'].values())[0]}, {list(kld[chr]['time_kld_per_thousand_operations'].values())[0]}, {list(kld[chr]['operations'].values())[0]}\n"
    with open('timing_sequential.txt', 'w') as fp:
        fp.write(timing)

    print("Done.")
