import os
import re
import numpy as np
from collections import Counter

import config


def transform_JASPAR_to_pfm(path_to_jaspar: str, out_path: str):
    with open(path_to_jaspar, "r") as jaspar:
        with open(out_path, "w") as out:
            for line in jaspar:
                if line.startswith(">"):
                    out.write(line)
                    counter = 0
                elif line.startswith(("A", "C", "G", "T")):
                    line = re.sub('A|G|C|T|\[|\]|\n', '', line).strip()
                    line = " ".join(line.split()) + "\n"
                    out.write(line)
                    counter += 1
                if counter == 4:
                    out.write("\n")
                    counter = 0




def read_pwms_from_pmf(path_to_pfm: str, pseudocount: float = 1.0) -> dict:
    """reads position count matrices from pmf file and transforms them to position wheight matrices

    Args:
        path_to_pfm (str): path to pmf file that should be read in
        pseudocount (float, optional): pseudocount, to prevent a devision by zero later. Defaults to 1.0.

    Returns:
        dict: dictionary with pwms and Metainformation. the rows represent A,C,G,T
    """


    pwms: dict = {}
    assert path_to_pfm[-4:] == ".pfm"
    assert os.path.exists(path_to_pfm)

    with open(path_to_pfm, 'r') as fp:
        line_counter = 0
        for line in fp:
            if line[0] == '>':  # header lines (with name and metainfo)
                name = line[1:].split()[0].strip()
                tf = line.split()[1].strip()
                pwms[name] = {'tf': tf, 'pwm': ''}
                line_counter = 1

            elif line_counter > 0:
                line_counter += 1
                pwms[name]['pwm'] = pwms[name]['pwm'] + line.replace('/n', ' ')

                if line_counter == 5:  # fourth line of array, but we added 1 before, so 5
                    if "." in pwms[name]['pwm']:  # Float values
                        new_pwm = np.fromstring(pwms[name]['pwm'], dtype=float, sep=' ').reshape((4, -1))
                    else: # int values
                        new_pwm = np.fromstring(pwms[name]['pwm'], dtype=int, sep=' ').reshape((4, -1))
                    new_pwm = new_pwm + pseudocount  # to prevent division by zero problems later (this is a standard procedure)
                    new_pwm = new_pwm / new_pwm.sum(axis=0)
                    pwms[name]['pwm'] = new_pwm
                    line_counter = 0

            else:  # empty line
                pass

    return pwms


def calculate_background(path: str = False) -> np.ndarray:
    """calculates background probabilities if given path. Defaults to the backgrounds probabilities of the human genome if no path given.

    Args:
        path (str, optional): Path to fasta file with dna sequence from wich to calculate background probability. Defaults to the backgrounds probabilities of the human genome if no path given
    Returns:
        background_probabilities (np.ndarray):  Array with background probabilities (order: A,C,G,T,N)
    """

    if not path:  # assume it is a human genome
        return np.array([0.28079571, 0.19395748, 0.19479377, 0.28169493, 0.0487581])  # ACGTN
    letter_counter = Counter()
    with open(path, 'r') as fp:
        for line in fp:
            if line[0] == ">":
                if "Primary Assembly" in line:
                    reading = True
                    print(line)
                    print()
                else:
                    reading = False
            elif reading:
                letter_counter.update(Counter(line.upper()))
            else:
                pass
    background = np.array(
        [letter_counter["A"], letter_counter["C"], letter_counter["G"], letter_counter["T"], letter_counter["N"]],
        dtype=np.float64)
    background = background / background.sum()
    return background

def sequence2nparray(sequence: str) -> np.ndarray:
    array = sequence.upper().replace("A", "0").replace("C", "1").replace("G", '2').replace("T", '3').replace("N", '4')
    array = np.array(list(array), dtype=np.uint8)
    return array

def extractSequenceWorker(path:str) -> dict:
    with open(path, 'r') as fp:
        text = fp.read()
    header, text = text.split('\n', maxsplit=1)
    text = text.replace("\n", "").upper()
    text = re.sub(r'[^AGCTN]', 'N', text)
    text = text.replace("A", "0").replace("C", "1").replace("G", '2').replace("T", '3').replace("N", '4')
    return_dict = {'header': header, 'sequence': np.array(list(text), dtype=np.uint8)}

    return return_dict



def splitGenome(path:str, files_path:str):
    print(path)
    print(files_path)
    with open(files_path+os.sep+'file_list.txt', "a+") as file_list:
        with open(path, 'r') as genome_file:
            writing = False
            for line in genome_file:
                if line.startswith(">"):
                    if "outfile" in locals():
                        outfile.close()
                    if "Primary Assembly" in line and "chromosome " in line and ">NC" in line:
                        chromosome_name = line[line.find("chromosome"):line.find("chromosome")+13].strip().replace(",", "").replace(" ", "_")
                        outfile = open(files_path+os.sep+chromosome_name+".fna", "a")
                        file_list.write(chromosome_name+".fna"+"\n")
                        writing = True
                        outfile.write(line)
                        print(line)
                    else:
                        writing = False
                elif writing:
                    outfile.write(line)
                else:
                    pass
    sortSplittedChromosomeFiles(files_path+os.sep+'file_list.txt')


def sortSplittedChromosomeFiles(path: str):
    """
    Sorts the files of the splitted chormosomes by file size
    Args:
        path: path to the file containing the information, in which files the genome was splitted

    Returns: nothing, list will be written back sorted

    """
    correct_files = []
    with open(path, 'r') as fp:
        for line in fp:
            correct_files.append(line.strip())

    dir = os.path.dirname(path)
    sorted_files_in_dir = sorted(os.listdir(dir),
           key=lambda x: os.stat(os.path.join(dir, x)).st_size,
           reverse=True)

    outstr = ""
    for file in sorted_files_in_dir:
        if file in correct_files:
            outstr += file+"\n"
    outstr = outstr[:-1]
    with open(path, "w") as fp:
        fp.write(outstr)
